@@ -10,7 +10,9 @@
 import Errors
 from Regexps import BOL, EOL, EOF
 
-class Scanner:
+import cython
+
+class Scanner(object):
   """
   A Scanner is used to read tokens from a stream of characters
   using the token set specified by a Plex.Lexicon.
@@ -42,23 +44,23 @@ class Scanner:
 
   """
 
-  lexicon = None        # Lexicon
-  stream = None         # file-like object
-  name = ''
-  buffer = ''
-  buf_start_pos = 0     # position in input of start of buffer
-  next_pos = 0          # position in input of next char to read
-  cur_pos = 0           # position in input of current char
-  cur_line = 1          # line number of current char
-  cur_line_start = 0    # position in input of start of current line
-  start_pos = 0         # position in input of start of token
-  start_line = 0        # line number of start of token
-  start_col = 0         # position in line of start of token
-  text = None           # text of last token read
-  initial_state = None  # Node
-  state_name = ''       # Name of initial state
-  queue = None          # list of tokens to be returned
-  trace = 0
+#  lexicon = None        # Lexicon
+#  stream = None         # file-like object
+#  name = ''
+#  buffer = ''
+#  buf_start_pos = 0     # position in input of start of buffer
+#  next_pos = 0          # position in input of next char to read
+#  cur_pos = 0           # position in input of current char
+#  cur_line = 1          # line number of current char
+#  cur_line_start = 0    # position in input of start of current line
+#  start_pos = 0         # position in input of start of token
+#  start_line = 0        # line number of start of token
+#  start_col = 0         # position in line of start of token
+#  text = None           # text of last token read
+#  initial_state = None  # Node
+#  state_name = ''       # Name of initial state
+#  queue = None          # list of tokens to be returned
+#  trace = 0
 
   def __init__(self, lexicon, stream, name = '', initial_pos = None):
     """
@@ -73,6 +75,19 @@ class Scanner:
       |name| is optional, and may be the name of the file being
       scanned or any other identifying string.
     """
+    self.trace = 0
+
+    self.buffer = ''
+    self.buf_start_pos = 0
+    self.next_pos = 0
+    self.cur_pos = 0
+    self.cur_line = 1
+    self.start_pos = 0
+    self.start_line = 0
+    self.start_col = 0
+    self.text = None
+    self.state_name = None
+    
     self.lexicon = lexicon
     self.stream = stream
     self.name = name
@@ -117,12 +132,8 @@ class Scanner:
     self.start_pos = self.cur_pos
     self.start_line = self.cur_line
     self.start_col = self.cur_pos - self.cur_line_start
-#		if self.trace:
-#			action = self.run_machine()
-#		else:
-#			action = self.run_machine_inlined()
     action = self.run_machine_inlined()
-    if action:
+    if action is not None:
       if self.trace:
         print("Scanner: read: Performing %s %d:%d" % (
           action, self.start_pos, self.cur_pos))
@@ -131,22 +142,12 @@ class Scanner:
       return (text, action)
     else:
       if self.cur_pos == self.start_pos:
-        if self.cur_char == EOL:
+        if self.cur_char is EOL:
           self.next_char()
-        if not self.cur_char or self.cur_char == EOF:
+        if self.cur_char is None or self.cur_char is EOF:
           return ('', None)
       raise Errors.UnrecognizedInput(self, self.state_name)
-  
-  def run_machine(self):
-    """
-    Run the machine until no more transitions are possible.
-    """
-    self.state = self.initial_state
-    self.backup_state = None
-    while self.transition():
-      pass
-    return self.back_up()
-  
+
   def run_machine_inlined(self):
     """
     Inlined version of run_machine for speed.
@@ -170,7 +171,7 @@ class Scanner:
       # Begin inlined self.save_for_backup()
       #action = state.action #@slow
       action = state['action'] #@fast
-      if action:
+      if action is not None:
         backup_state = (
           action, cur_pos, cur_line, cur_line_start, cur_char, input_state, next_pos)
       # End inlined self.save_for_backup()
@@ -232,7 +233,7 @@ class Scanner:
         if trace: #TRACE#
           print("blocked")  #TRACE#
         # Begin inlined: action = self.back_up()
-        if backup_state:
+        if backup_state is not None:
           (action, cur_pos, cur_line, cur_line_start, 
             cur_char, input_state, next_pos) = backup_state
         else:
@@ -244,49 +245,12 @@ class Scanner:
     self.cur_line_start = cur_line_start
     self.cur_char = cur_char
     self.input_state = input_state
-    self.next_pos	 = next_pos
+    self.next_pos     = next_pos
     if trace: #TRACE#
-      if action: #TRACE#
-        print("Doing " + action) #TRACE#
+      if action is not None: #TRACE#
+        print("Doing %s" % action) #TRACE#
     return action
-    
-#	def transition(self):
-#		self.save_for_backup()
-#		c = self.cur_char
-#		new_state = self.state.new_state(c)
-#		if new_state:
-#			if self.trace:
-#				print "Scanner: read: State %d: %s --> State %d" % (
-#					self.state.number, repr(c), new_state.number)
-#			self.state = new_state
-#			self.next_char()
-#			return 1
-#		else:
-#			if self.trace:
-#				print "Scanner: read: State %d: %s --> blocked" % (
-#					self.state.number, repr(c))
-#			return 0
-  
-#	def save_for_backup(self):
-#		action = self.state.get_action()
-#		if action:
-#			if self.trace:
-#				print "Scanner: read: Saving backup point at", self.cur_pos
-#			self.backup_state = (
-#				action, self.cur_pos, self.cur_line, self.cur_line_start, 
-#				self.cur_char, self.input_state, self.next_pos)
-  
-#	def back_up(self):
-#		backup_state = self.backup_state
-#		if backup_state:
-#			(action, self.cur_pos, self.cur_line, self.cur_line_start, 
-#				self.cur_char, self.input_state, self.next_pos) = backup_state
-#			if self.trace:
-#				print "Scanner: read: Backing up to", self.cur_pos
-#			return action
-#		else:
-#			return None
-  
+
   def next_char(self):
     input_state = self.input_state
     if self.trace:
@@ -317,26 +281,7 @@ class Scanner:
       self.cur_char = ''
     if self.trace:
       print("--> [%d] %d %s" % (input_state, self.cur_pos, repr(self.cur_char)))
-    
-#	def read_char(self):
-#		"""
-#    Get the next input character, filling the buffer if necessary.
-#    Returns '' at end of file.
-#    """
-#		next_pos = self.next_pos
-#		buf_index = next_pos - self.buf_start_pos
-#		if buf_index == len(self.buffer):
-#			discard = self.start_pos - self.buf_start_pos
-#			data = self.stream.read(0x1000)
-#			self.buffer = self.buffer[discard:] + data
-#			self.buf_start_pos = self.buf_start_pos + discard
-#			buf_index = buf_index - discard
-#			if not data:
-#				return ''
-#		c = self.buffer[buf_index]
-#		self.next_pos = next_pos + 1
-#		return c
-  
+
   def position(self):
     """
     Return a tuple (name, line, col) representing the location of
@@ -374,6 +319,3 @@ class Scanner:
     Override this method if you want something to be done at
     end of file.
     """
-
-# For backward compatibility:
-setattr(Scanner, "yield", Scanner.produce)
